C   circles - calculate identified circles statistics from CMB data (cosmic topology)
C
C   Copyright (C) 2004 Boud Roukema
C
C   This program is free software; you can redistribute it and/or modify
C   it under the terms of the GNU General Public License as published by
C   the Free Software Foundation; either version 2, or (at your option)
C   any later version.
C
C   This program is distributed in the hope that it will be useful,
C   but WITHOUT ANY WARRANTY; without even the implied warranty of
C   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C   GNU General Public License for more details.
C
C   You should have received a copy of the GNU General Public License
C   along with this program; if not, write to the Free Software Foundation,
C   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
C

C     ==== to read the WMAP linear combination data file ====
      subroutine rdWMAP(npix,indir2,nindir,
     _     temp,errtemp,gallong,gallat,tempcp)


C     INPUTS:
      character*256  indir2    ! directory with input files
      integer        nindir    ! number of characters in  indir

C     INTERNAL:
      real*8   pi
      parameter(pi=3.1415926535898)

c      parameter(npix=6144,maxn=8096) !! WARNING declared more than once
c      parameter(npix=6120,maxn=8096) !! WARNING declared more than once

c     -> now done in pxallo.c with declarations in circles.h
c
c      include 'npix.dec'

c      integer*4  ipix(maxn)
c     ,nobspix(maxn)

c     -> now done in pxallo.c with declarations in circles.h
c
c      real*4   temp(maxn),errtemp(maxn) !   Temperature and uncert.
c      real*4   gallong(maxn),gallat(maxn) ! galactic
      real*4   temp(*),errtemp(*) !   Temperature and uncert.
      real*4   gallong(*),gallat(*) ! galactic

      real*4   tempcp(*) ! temporary copy of temp(*)

c      real*4   eclong(maxn),eclat(maxn) ! ecliptic
c      real*4   eqlong(maxn),eqlat(maxn) ! ra dec in degrees, degrees

c      real*4   sortcrit(2,maxn)  !! wrong order !!
c      real*4   sortcrit(maxn,2)

c      parameter(nxint=180,nyint=90)
      include 'nxint.dec'

      integer*4 inear(nxint,nyint)    ! nearest pixel number
      real*4    dinear(nxint,nyint)
      real*4    dxint,dyint
c     -- Eq. Cartesian coords of table --
      real*4    x(nxint,nyint),y(nxint,nyint),z(nxint,nyint) 

      logical*4 measur,noise      ! measured or random; (signal or) noise
c      real*4    ranerr(maxn)


c     ---- convert pixel number to  angle using isolat package ----
c     e.g. http://cosmo.torun.pl/GPLdownload/dodec/isolat-0.1.11.tar.gz

      parameter(n_res=512) ! hardwired for WMAP
      parameter(inorth= 0) !  multiple of pi/2 of North pole = -1,0,1 or 2 
      parameter(ifpixel = 1)    ! first pixel number 
      parameter(j_warn = 0)  ! j_warn = 1 turns off warnings
      real*8          theta,phi

c      logical*4       doerrupper

C     EXTERNAL:
      integer  pxstat  ! pxstat.c

C     COMMON:
      common /noisecomm/ measur,noise

c      common /wmapcomm/  temp,errtemp,gallong,gallat, eqlong,eqlat,

c      common /wmapcomm/  temp,errtemp,gallong,gallat, 
c     _     dsm,gb0,dxint,dyint,inear,dinear,x,y,z  ,ranerr
      common /wmapcomm/ npixcp,
     _     dsm,gb0,dxint,dyint,inear,dinear,x,y,z

c      common /erruppcomm/ doerrupper,errupper ! shared with plotdodec (at least)
      include 'errupp.dec'

      character*256  cmbfile

c      byte      by_array(40)
      integer*4 intarray(20)
      real*4    r__array(20)
      equivalence(intarray(1),r__array(1))

      integer*4  iswbyte
      external   iswbyte


      logical*4  domask
      logical*4  test

      integer*4  iwarn1  ! read file error
      parameter(mxwar1= 10) ! max no of warnings type 1

      integer    iwverb  ! verbose interaction ?

      test= .false. !.true. !

      domask= .false.              !.true. !

      iwverb=1 

      dsm= 3.5*pi/180.0

c     ==== copy npix to common block ====
      npixcp=npix
      
      print*,'rdWMAP beginning: npix=',npix

c     ==== check that memory has been allocated ====
      i= pxstat()
      if(i.ne.0)then
         print*,'rdWMAP WARNING: pxstat returned ',i,
     _        ' - there is probably a memory allocation problem.'
         print*,'Did you forget to call pxallo?'
      endif
      if(test)then
         print*,'rdWMAP: before print*,temp(1), temp(1)= 4.3'
         print*,'temp(1)=',temp(1)
         temp(1)=4.3
         print*,'rdWMAP: after temp(1)= ',temp(1)
      endif


c     ==== read in file ====

      iunit= 1
      irwmode=0

      if(nindir.gt.0.and.indir2(nindir:nindir).ne.'/')then
         cmbfile=indir2(1:nindir)//'/'//'map_ilc_yr1_v1.fits'
      else
         cmbfile=indir2(1:min(nindir,256))//'map_ilc_yr1_v1.fits' ! kaz
      endif

      istatus=0
      print*,'istatus=',istatus
      print*,'Will read :'
      print*,cmbfile
      call ftopen(iunit,cmbfile,irwmode,iblocksize,istatus)
      print*,'irwmode,iblocksize=',irwmode,iblocksize
      print*,'istatus=',istatus

      nhdu=2
      ihdutype=2
      call ftmahd(iunit,nhdu,ihdutype,istatus)
      print*,'after ftmahd: istatus=',istatus

      n_long=8

c      errupper= 10.0e-3 ! mK - estimate of upper limit to error /pixel 

      iwarn1= 0

      do i=1,npix
         if(test)print*,'i=',i
         call ftgtbb(iunit,i,1,n_long,intarray,istatus)
         if(istatus.ne.0.and.iwarn1.lt.mxwar1)then
            print*,'rdWMAP WARNING readerror',
     _           'i,istatus=',i,istatus
            print*,'probably something is wrong with the',
     _           ' following file: ',
     _           cmbfile
            iwarn1= iwarn1+1
         endif
         if(test)print*,'rdWMAP: after a call to ftgtbb '

         do k=1,n_long/4
            intarray(k)=iswbyte(intarray(k)) !intarray(k)  !
         enddo

         if(test)print*,'rdWMAP: after iswbyte line'

c         ipix(i)=i
         if(test)print*,'i,temp(i)=',i,temp(i)

         temp(i)=r__array(1)
         if(test)print*,'after temp assignment: i,temp(i)=',i,temp(i)


         if(test)print*,'rdWMAP: after an evaluation to temp(1)'

c         nobspix(i)=r__array(2) ! this is 'counts ' - is this no of obsvns??
         ncount=r__array(2) ! this is 'counts ' - is this no of obsvns??
         if(doerrupper)then
            errtemp(i)= errupper
         else
            stop 'rdWMAP: no alternative to errupper defined.'
         endif

         if(domask)then
c            temp(i)=nobspix(i)  ! = 1.0 or 0.0
            temp(i)=ncount  ! = 1.0 or 0.0
         endif

c     --- call isolat routine to convert pixel number to angle on sky  ---
         call getang(n_res, inorth, ifpixel, j_warn, i,
     _        phi,theta, nr_out)

         if(j_warn.eq.0.and.nr_out.ne.n_res)then
            print*,'rdWMAP: WARNING: n_res .ne. nr_out = ',
     _           n_res,nr_out
         endif

         gallong(i)= phi/(pi/180.0)
c         gallat(i)= theta/(pi/180.0) -90.0 
         gallat(i)= -(theta/(pi/180.0) -90.0) ! 24.04.2003

c     --- test only ---
c         if(abs(gallat(i)).gt.88.0.or.i.lt.5)print*,
c     _        'i,gallat(i),gallong(i)=',
c     _        i,gallat(i),gallong(i)

c         eclong(i)=r__array(5)
c         eclat(i)=r__array(6)
c         gallong(i)=r__array(7)
c         gallat(i)=r__array(8)
c         eqlong(i)=r__array(9)
c         eqlat(i)=r__array(10)

c         print*,'ipix(i),temp(i),nobspix(i),errtemp(i)='
c         print*,ipix(i),temp(i),nobspix(i),errtemp(i)

c     disabled in 0.1.12 -> add back later as option
c         ranerr(i)=gauss0(0.0,errtemp(i)) ! generate errors from errtemp
      enddo
      

      call ftclos(iunit,istatus)
      if(istatus.ne.0)print*,'i,istatus=',i,istatus

      print*,'WMAP sample output: '
c      print*,'(ipix(i),temp(i),errtemp(i),gallong(i),gallat(i),i=1,3)'
      print*,'(i,temp(i),errtemp(i),gallong(i),gallat(i),i=1,3)'
      do i=1,5
c         print*,ipix(i),temp(i),errtemp(i),gallong(i),gallat(i)
         print*,i,temp(i),errtemp(i),gallong(i),gallat(i)
      enddo

c     ==== initialise tables ====

c     ---- make gl-gb tables of closest pixel and dist to cl pixel ----
      dyint=180.0/real(nyint)  ! degrees /table unit
      dxint=360.0/real(nxint)  ! degrees /table unit
      print*,'dxint,dyint=',dxint,dyint

      gb0=-90.0
 
      if(test)print*,'rdWMAP: nxint,nyint=',nxint,nyint
      do j=1,nyint
         gb= gb0+ (real(j)-0.5)*dyint
         do i=1,nxint
            if(test.and.i.eq.152.and.j.eq.59)print*,
     _           'rdWMAP test 1 '
            inear(i,j)=-99
            dinear(i,j)=9e9
            gl= (real(i)-0.5)*dxint
            x(i,j)= cos( gb*pi/180.0 )*cos(gl*pi/180.0)
            y(i,j)= cos( gb*pi/180.0 )*sin(gl*pi/180.0)
            z(i,j)= sin( gb*pi/180.0 )
            if(test.and.i.eq.152.and.j.eq.59)print*,
     _           'rdWMAP test 2 '
         enddo
      enddo


      measur=.true.
      noise=.false.
      print*,'wmaplin(10.0,0.0),wmaplin(180.0,89.0)='
      xx= wmaplin(10.0,0.0,
     _     temp,errtemp,gallong,gallat)
      yy=wmaplin(180.0,89.0,
     _     temp,errtemp,gallong,gallat)
      print*,xx,yy

      return
      end 


      real*4   function wmaprot(gall,galb)
c     ==== gives wmap temp fluctuation at gall, galb,
c     after a rotation from (glrot,gbrot) to the "north pole" ====
c     12.01.2004

      common /rotcomm/ glrot0,gbrot0

      call nprotate(glrot0,gbrot0, gall,galb, glout,gbout)
      wmaprot=wmaplin(glout,gbout,
     _     temp,errtemp,gallong,gallat)

      return
      end




      real*4   function  wmaplin(gall,galb,
     _     temp,errtemp,gallong,gallat)
c     ==== output may be either temperature variations or their 
c     uncertainties, depending on whether .not.(noise) or (noise) 
c     is set respectively ====
c     -- the  /noisecomm/  common block should be declared in the 
c     calling subroutine(s).

C     INPUTS:
      real*4   gall,galb        ! galactic longitude, latitude input

c      parameter(npix=6144,maxn=8096)  !! WARNING declared several times
c      parameter(npix=6120,maxn=8096)  !! WARNING declared several times
c      include 'npix.dec'

c      real*4   temp(maxn),errtemp(maxn) !   Temperature and uncert.
c      real*4   gallong(maxn),gallat(maxn) ! galactic
      real*4   temp(*),errtemp(*) !   Temperature and uncert.
      real*4   gallong(*),gallat(*) ! galactic


c      real*4   eclong(maxn),eclat(maxn) ! ecliptic
c      real*4   eqlong(maxn),eqlat(maxn) ! ra dec in degrees, degrees


C     EXTERNAL:
      integer  pxstat  ! pxstat.c


C     INTERNAL:
      real*8    pid,twopid
      parameter(pi=3.1415926535898)
      parameter(pid=3.1415926535898d0,twopid=2.0d0*pid)


c      parameter(nxint=180,nyint=90)
      include 'nxint.dec'

      integer*4 inear(nxint,nyint)    ! nearest pixel number
      real*4    dinear(nxint,nyint)
      real*4    dxint,dyint
c     -- Eq. Cartesian coords of table --
      real*4    x(nxint,nyint),y(nxint,nyint),z(nxint,nyint) 

      logical*4 measur,noise    

c     disabled in 0.1.12 -> add back later as option
c      real*4    ranerr(maxn)

c     ---- convert pixel number to  angle using isolat package ----
c     e.g. http://cosmo.torun.pl/GPLdownload/dodec/isolat-0.1.11.tar.gz

      parameter(n_res=512) ! hardwired for WMAP
      parameter(inorth= 0) !  multiple of pi/2 of North pole = -1,0,1 or 2 
      parameter(ifpixel = 1)    ! first pixel number 
      parameter(j_warn = 0)  ! j_warn = 1 turns off warnings
      real*8          theta,phi, told
      parameter(told=1e-6)

      common /noisecomm/ measur,noise

c      common /wmapcomm/  temp,errtemp,gallong,gallat,eqlong,eqlat,
      common /wmapcomm/ npixcp,
     _     dsm,gb0,dxint,dyint,inear,dinear,x,y,z 

c      print*,'BEGINNING of  wmaplin(...)',
c     _     'measur,noise=',measur,noise

      npix=npixcp
c      print*,'rdWMAP beginning: npix=',npix
      i= pxstat()
      if(i.ne.0)then
         print*,'rdWMAP WARNING: pxstat returned ',i,
     _        ' - there is probably a memory allocation problem.'
         print*,'Did you forget to call pxallo?'
      endif


      theta=((-galb)+ 90.0) *pi/180.0 
      phi= gall*pi/180.0

      theta=min(max(told,theta),pid) ! getpix does not check limits
      phi=min(max(told,phi),twopid)

      call getpix(n_res, inorth, ifpixel, j_warn, phi,theta,
     _     ipix, nr_out)

c      xx= cos( galb*pi/180.0 )*cos(gall*pi/180.0)
c      yy= cos( galb*pi/180.0 )*sin(gall*pi/180.0)
c      zz= sin( galb*pi/180.0 )

      
      if(measur)then
         if(.not.(noise))then
            wmaplin=temp(ipix)
         else
            wmaplin=errtemp(ipix)
         endif
      elseif(noise)then
         print*,'The noise option was disabled in circles-0.1.12,'
         print*,'and could be added in as an option in a later'
         print*,'version. This code is GPL - please feel free'
         print*,'to update the package and redistribute the new'
         print*,'version according to the terms of the GPL. :)'
c     disabled in 0.1.12 -> add back later as option
c         wmaplin=ranerr(ipix)
      else
         print*,'measur,noise=',measur,noise,' not useful.'
         wmaplin=-99.9
      endif

      return
      end


      subroutine  smooWMAP(new,sigdeg,resdeg, indir2,nindir,
     _     temp,errtemp,gallong,gallat,tempcp)
c     !! modifies (at least) the array temp

c     INPUTS:
      logical*4       new ! is it a newly smoothed file?
      real*4          sigdeg ! # 1 sigma width in degrees
      real*4          resdeg    ! # resolution of smoothing in degrees


      character*256  indir2    ! directory with input files
      integer        nindir    ! number of characters in  indir

      parameter(pi=3.1415926535898)
      parameter(pi_180=pi/180.0)

c      include 'npix.dec'

      character*256   smoofile ! for reading in 

c      real*4   temp(maxn),errtemp(maxn) !   Temperature and uncert.
c      real*4   gallong(maxn),gallat(maxn) ! galactic
c      real*4   eqlong(maxn),eqlat(maxn) ! ra dec in degrees, degrees

      real*4   temp(*),errtemp(*) !   Temperature and uncert.
      real*4   gallong(*),gallat(*) ! galactic

c      real*4   tempcp(maxn) ! copy of temp while working on it
      real*4   tempcp(*) ! copy of temp while working on it

      include 'nxint.dec'

      integer*4 inear(nxint,nyint)    ! nearest pixel number
      real*4    dinear(nxint,nyint)
      real*4    dxint,dyint
c     -- Eq. Cartesian coords of table --
      real*4    x(nxint,nyint),y(nxint,nyint),z(nxint,nyint) 

      logical*4 measur,noise    
c     disabled in 0.1.12 -> add back later as option
c      real*4    ranerr(maxn)

      parameter(n_res=512) ! hardwired for WMAP
      parameter(inorth= 0) !  multiple of pi/2 of North pole = -1,0,1 or 2 
      parameter(ifpixel = 1)    ! first pixel number 
      parameter(j_warn = 0)  ! j_warn = 1 turns off warnings
      real*8          theta,phi

      external        wmaplin
      real*4          wmaplin

C     EXTERNAL:
      integer  pxstat  ! pxstat.c

      include 'errupp.dec'

      character*17    ch17

      real*8          zp8,rr8,dd8,aa8

      integer         iwverb

      common /noisecomm/ measur,noise

c      common /wmapcomm/  temp,errtemp,gallong,gallat,eqlong,eqlat,
      common /wmapcomm/ npixcp,
     _     dsm,gb0,dxint,dyint,inear,dinear,x,y,z 


      iwverb= 1  ! verbose


      npix=npixcp
      i= pxstat()
      if(i.ne.0)then
         print*,'smooWMAP WARNING: pxstat returned ',i,
     _        ' - there is probably a memory allocation problem.'
         print*,'Did you forget to call pxallo?'
      endif
      
c     --- allocate memory for tempcp(*) ----
c      i= pxal_2(iwverb,tempcp)
c      if(i.ne.0)then
c         print*,'smooWMAP WARNING: pxal_2 returned ',i,
c     _        ' - there is probably a memory allocation problem.'
c      endif
      


      if(nindir.gt.0.and.indir2(nindir:nindir).ne.'/')then
         smoofile=indir2(1:nindir)//'/'//'smooW1d.dat' 
      else
         smoofile=indir2(1:min(nindir,256))//'smooW1d.dat'
      endif

c      smoofile='/scratch0/boud/WMAP/smooW5d.dat'
c      smoofile='/scratch2/boud/WMAP/smooW5d.dat'

c      print*,'BEGINNING of  wmaplin(...)',
c     _     'measur,noise=',measur,noise

c     --- some hardwired warnings on reasonable smoothing parameters ---
      if(sigdeg.gt.10.0.or.resdeg.lt.0.1)then
         print*,'smooWMAP WARNING: sigdeg, resdeg=',sigdeg,resdeg
      endif

      tol=1e-4
      if(new)then  ! if not yet smoothed

c     ---- loop through all pixels ----
         sig2 = 2.0*sigdeg*sigdeg ! degrees
         sig2rad = sig2 *pi_180 *pi_180 ! radians
         resrad= resdeg*pi_180
         res2 = resdeg*resdeg

c     do ipix=1,npix            ! loop through 3M pixels
         ipix1=1
         ipix2=npix

c     ---- test only ----
         ihigh = 261600! 2097153        !523744         !
c         ipix1=ihigh-30000
c         ipix2=ihigh
c     ----

         do ipix=ipix1,ipix2 
            gall=gallong(ipix)
            galb=gallat(ipix)

c     ---- if not close to pole, then use  (b,l/cos(b))  smoothing ----
c            if(abs(galb).lt.90.0-40.0*sigdeg)then !! test only !!

            if(abs(galb).lt.90.0-4.0*sigdeg)then
c               print*,'smooWMAP TEST - non-pole calc, ipix=',ipix
               cosb= cos(galb*pi_180)
               cosb2= cosb*cosb
               nb= nint(2.0*sigdeg/resdeg)+1 ! go to 2 sigma
               nl= nb
               
               tt=0.0           ! temperature
               ww=0.0           ! weight
               do i=-nb,nb
                  gb= galb + real(i)*resdeg
                  do j=-nl,nl
                     gl= mod(gall + real(j)*resdeg/cosb + 720.0,
     _                    360.0)
                     if(gl.eq.0.0)gl=resdeg*0.5
                     
c                     aa2= (gl-gall)*(gl-gall)*cosb2 +    ! risk of 360 edge
c     _                    (gb-galb)*(gb-galb) ! angle squared from pixel

                     aa= real(i)
                     bb= real(j)
                     aa2= (aa*aa +bb*bb)*res2
                     dw = exp(-aa2/sig2)
                     ww= ww + dw
                     tt= tt + dw * wmaplin(gl,gb,
     _                    temp,errtemp,gallong,gallat)
                  enddo
               enddo
               if(ww.lt.tol)then
                  print*,'smooWMAP WARNING: ipix,gall,galb,ww=',
     _                 ipix,gall,galb,ww
               endif
               tempcp(ipix) = tt/ww
            else
c               print*,'smooWMAP TEST - pole calculation, ipix=',ipix
c     ---- if close to pole, use X-Y projection ---
               xp= cos( galb*pi_180 )*cos(gall*pi_180) ! radians
               yp= cos( galb*pi_180 )*sin(gall*pi_180) ! radians
               zp= sin( galb*pi_180 ) ! radians
               zp8= dble(zp)
               
               nx= nint(2.0*sigdeg/resdeg) +1 ! go to 2 sigma
               ny= nx
               
               tt=0.0           ! temperature
               ww=0.0           ! weight
               do i=-nx,nx
                  xx= xp + real(i)*resrad  ! radians
                  do j=-ny,ny
                     yy= yp + real(j)*resrad  ! radians
c                     call xyz_rda(xx,yy,zp, ! use fixed z of the pixel
c     _                    rr,dd,aa)
c                     call xyz_rda8(dble(xx),dble(yy),zp8,
                     call xyz_8rda(dble(xx),dble(yy),zp8,
     _                    rr8,dd8,aa8)
                     gb= real(dd8)
                     gl= real(aa8*15.0d0) ! hours -> degrees

c                     print*,'i,j,gb,gl=',i,j,gb,gl !! test only !!
                     if(gl.eq.0.0)gl=resdeg*0.5

                     if(gb.gt.90.0-tol)gb=90.0-tol
                     if(gb.lt.-90.0+tol)gb=-90.0+tol
                     
c                     aa2= (gl-gall)*(gl-gall) +   
c     _                    (gb-galb)*(gb-galb) ! angle squared from pixel
                     aa2=  (xp-xx)*(xp-xx) + (yp-yy)*(yp-yy) ! radians

                     dw= exp(-aa2/sig2rad)
                     ww= ww + dw
                     tt= tt + dw * wmaplin(gl,gb,
     _                    temp,errtemp,gallong,gallat)
                  enddo
               enddo
               if(ww.lt.tol)then
                  print*,'smooWMAP WARNING: ipix,gall,galb,ww=',
     _                 ipix,gall,galb,ww
               endif
               tempcp(ipix) = tt/ww
            endif
            
         enddo                  !       do ipix=1,npix           

c     --- copy to temp() ----
         do ipix=ipix1,ipix2 ! 1,npix
            temp(ipix)=tempcp(ipix)  
         enddo

c     --- write out ----
         open(1,file='smooWMAP.dat',access='sequential',
     _        form='unformatted',status='unknown')
         write(1)'smoothed WMAP map'
         write(1)sigdeg,resdeg

c         write(1)temp
c     -> changed in version 0.1.12 -> memory allocated by pxallo, not by fortran
         write(1)(temp(i),i=1,npix)

         close(1)

      else! if the file is claimed to already exist
c         open(1,file='smooWMAP.dat',access='sequential',
         open(1,file=smoofile,access='sequential',
     _           form='unformatted',status='old')
         read(1)ch17
         if(ch17.ne.'smoothed WMAP map')
     _        print*,'WARNING: strange file!'

         read(1)sigd,resd
c     -> changed in version 0.1.12 -> memory allocated by pxallo, not by fortran
c         read(1)temp
         read(1)(temp(i),i=1,npix)

         close(1)

         if(abs((sigd-sigdeg)/sigdeg).gt.tol.or.
     _        abs((resd-resdeg)/resdeg).gt.tol)then
            print*,'WARNING: probably wrong file, ',
     _           'sigdeg,resdeg,sigd,resd=',
     _           sigdeg,resdeg,sigd,resd
         endif
      endif



c     --- free up memory for tempcp(*) ----
c      i= pxfr_2(iwverb)
c      if(i.ne.-1)then   ! -1 means memory freed up or no allocation attempt
c         print*,'smooWMAP WARNING: pxal_2 returned ',i,
c     _        ' - there was a problem freeing up memory :( '
c      endif
      

      return
      end

